// bbkeys.cc for blackbox
// ------ note from the Gee-I'm-Sorry-The-Code-Looks-Ugly department-----
// ------ for vi users-- ":set ts=2"  =:)
//
//  Copyright (c) 1999-2000 by Jason Kasper j_dot_kasper@usa.net
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// (See the included file COPYING / GPL-2.0)
//

#ifdef    HAVE_CONFIG_H
#  include "config.h"
#endif

#ifdef    HAVE_STDIO_H
#  include <stdio.h>
#endif													// HAVE_STDIO_H

#ifdef    HAVE_CTYPE_H
#  include <ctype.h>
#endif													// HAVE_CTYPE_H

#ifdef    STDC_HEADERS
#  include <stdlib.h>
#  include <string.h>
#endif													// STDC_HEADERS

#ifdef    HAVE_SYS_TYPES_H
#  include <sys/types.h>
#endif													// HAVE_SYS_TYPES_H

#ifdef    HAVE_SYS_WAIT_H
#  include <sys/wait.h>
#endif													// HAVE_SYS_WAIT_H

#ifdef    HAVE_ERRNO_H
#  include <errno.h>
#endif													// HAVE_ERRNO_H

#ifdef    HAVE_UNISTD_H
#  include <unistd.h>
#endif													// HAVE_UNISTD_H

#ifdef    HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif													// HAVE_SYS_STAT_H

#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/keysym.h>

#include "bbkeys.hh"
#include "version.h"
#include "Timer.hh"


/*--------------------------------------------------------------------*/

/* Ripped shamelessly from wmaker's xmodifier.c--converted for our use*/

/*--------------------------------------------------------------------*/

/* Grok X modifier mappings for shortcuts.
   
   Most of this code was taken from src/event-Xt.c in XEmacs 20.3-b17.
   The copyright(s) from the original XEmacs code are included below.

   Perpetrator: Sudish Joseph <sj@eng.mindspring.net>, Sept. 1997. */

/* The event_stream interface for X11 with Xt, and/or tty frames.
   Copyright (C) 1991, 1992, 1993, 1994, 1995 Free Software Foundation, Inc.
   Copyright (C) 1995 Sun Microsystems, Inc.
   Copyright (C) 1996 Ben Wing.

This file is part of XEmacs.

XEmacs is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

XEmacs is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with XEmacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/************************************************************************/

/*                            keymap handling                           */

/************************************************************************/

/* X bogusly doesn't define the interpretations of any bits besides
   ModControl, ModShift, and ModLock; so the Interclient Communication
   Conventions Manual says that we have to bend over backwards to figure
   out what the other modifier bits mean.  According to ICCCM:

   - Any keycode which is assigned ModControl is a "control" key.

   - Any modifier bit which is assigned to a keycode which generates Meta_L
     or Meta_R is the modifier bit meaning "meta".  Likewise for Super, Hyper,
     etc.

   - Any keypress event which contains ModControl in its state should be
     interpreted as a "control" character.

   - Any keypress event which contains a modifier bit in its state which is
     generated by a keycode whose corresponding keysym is Meta_L or Meta_R
     should be interpreted as a "meta" character.  Likewise for Super, Hyper,
     etc.

   - It is illegal for a keysym to be associated with more than one modifier
     bit.

   This means that the only thing that emacs can reasonably interpret as a
   "meta" key is a key whose keysym is Meta_L or Meta_R, and which generates
   one of the modifier bits Mod1-Mod5.

   Unfortunately, many keyboards don't have Meta keys in their default
   configuration.  So, if there are no Meta keys, but there are "Alt" keys,
   emacs will interpret Alt as Meta.  If there are both Meta and Alt keys,
   then the Meta keys mean "Meta", and the Alt keys mean "Alt" (it used to
   mean "Symbol," but that just confused the hell out of way too many people).

   This works with the default configurations of the 19 keyboard-types I've
   checked.

   Emacs detects keyboard configurations which violate the above rules, and
   prints an error message on the standard-error-output.  (Perhaps it should
   use a pop-up-window instead.)
 */

static int MetaMask, HyperMask, SuperMask, AltMask, ModeMask;

char *ToolWindow::index_to_name(int indice)
{
	switch (indice) {
	case ShiftMapIndex:
		return "ModShift";
	case LockMapIndex:
		return "ModLock";
	case ControlMapIndex:
		return "ModControl";
	case Mod1MapIndex:
		return "Mod1";
	case Mod2MapIndex:
		return "Mod2";
	case Mod3MapIndex:
		return "Mod3";
	case Mod4MapIndex:
		return "Mod4";
	case Mod5MapIndex:
		return "Mod5";
	default:
		return "???";
	}
}

void ToolWindow::x_reset_modifier_mapping(Display * display)
{
	int modifier_index, modifier_key, column, mkpm;
	int warned_about_overlapping_modifiers = 0;
	int warned_about_predefined_modifiers = 0;
	int warned_about_duplicate_modifiers = 0;
	int meta_bit = 0;
	int hyper_bit = 0;
	int super_bit = 0;
	int alt_bit = 0;
	int mode_bit = 0;
	XModifierKeymap *x_modifier_keymap = XGetModifierMapping(display);

#define modwarn(name,old,other)						    \
  warned_about_overlapping_modifiers = 1

#define modbarf(name,other)						    \
  warned_about_predefined_modifiers = 1

#define check_modifier(name,mask)					    \
    warned_about_predefined_modifiers = 1

#define store_modifier(name,old)					   \
  if (old && old != modifier_index)					   \
    warned_about_duplicate_modifiers = 1;				   \
  if (modifier_index == ShiftMapIndex) modbarf (name,"ModShift");	   \
  else if (modifier_index == LockMapIndex) modbarf (name,"ModLock");	   \
  else if (modifier_index == ControlMapIndex) modbarf (name,"ModControl"); \
  else if (sym == XK_Mode_switch)					   \
    mode_bit = modifier_index; /* Mode_switch is special, see below... */  \
  else if (modifier_index == meta_bit && old != meta_bit)		   \
    modwarn (name, meta_bit, "Meta");					   \
  else if (modifier_index == super_bit && old != super_bit)		   \
    modwarn (name, super_bit, "Super");					   \
  else if (modifier_index == hyper_bit && old != hyper_bit)		   \
    modwarn (name, hyper_bit, "Hyper");					   \
  else if (modifier_index == alt_bit && old != alt_bit)			   \
    modwarn (name, alt_bit, "Alt");					   \
  else									   \
    old = modifier_index;

	mkpm = x_modifier_keymap->max_keypermod;
	for (modifier_index = 0; modifier_index < 8; modifier_index++)
		for (modifier_key = 0; modifier_key < mkpm; modifier_key++) {
			KeySym last_sym = 0;
			for (column = 0; column < 4; column += 2) {
				KeyCode code =
					 x_modifier_keymap->modifiermap[modifier_index * mkpm +
																					modifier_key];
				KeySym sym = (code ? XKeycodeToKeysym(display, code, column) : 0);
				if (sym == last_sym)
					continue;
				last_sym = sym;
				switch (sym) {
				case XK_Mode_switch:
					store_modifier("Mode_switch", mode_bit);
					break;
				case XK_Meta_L:
					store_modifier("Meta_L", meta_bit);
					break;
				case XK_Meta_R:
					store_modifier("Meta_R", meta_bit);
					break;
				case XK_Super_L:
					store_modifier("Super_L", super_bit);
					break;
				case XK_Super_R:
					store_modifier("Super_R", super_bit);
					break;
				case XK_Hyper_L:
					store_modifier("Hyper_L", hyper_bit);
					break;
				case XK_Hyper_R:
					store_modifier("Hyper_R", hyper_bit);
					break;
				case XK_Alt_L:
					store_modifier("Alt_L", alt_bit);
					break;
				case XK_Alt_R:
					store_modifier("Alt_R", alt_bit);
					break;
				case XK_Control_L:
					check_modifier("Control_L", ControlMask);
					break;
				case XK_Control_R:
					check_modifier("Control_R", ControlMask);
					break;
				case XK_Shift_L:
					check_modifier("Shift_L", ShiftMask);
					break;
				case XK_Shift_R:
					check_modifier("Shift_R", ShiftMask);
					break;
				case XK_Shift_Lock:
					check_modifier("Shift_Lock", LockMask);
					break;
				case XK_Caps_Lock:
					check_modifier("Caps_Lock", LockMask);
					break;

					/* It probably doesn't make any sense for a modifier bit to be
					   assigned to a key that is not one of the above, but OpenWindows
					   assigns modifier bits to a couple of random function keys for
					   no reason that I can discern, so printing a warning here would
					   be annoying. */
				}
			}
		}
#undef store_modifier
#undef check_modifier
#undef modwarn
#undef modbarf

	/* If there was no Meta key, then try using the Alt key instead.
	   If there is both a Meta key and an Alt key, then the Alt key
	   is not disturbed and remains an Alt key. */
	if (!meta_bit && alt_bit)
		meta_bit = alt_bit, alt_bit = 0;

	/* mode_bit overrides everything, since it's processed down inside of
	   XLookupString() instead of by us.  If Meta and Mode_switch both
	   generate the same modifier bit (which is an error), then we don't
	   interpret that bit as Meta, because we can't make XLookupString()
	   not interpret it as Mode_switch; and interpreting it as both would
	   be totally wrong. */
	if (mode_bit) {
		const char *warn = 0;
		if (mode_bit == meta_bit)
			warn = "Meta", meta_bit = 0;
		else if (mode_bit == hyper_bit)
			warn = "Hyper", hyper_bit = 0;
		else if (mode_bit == super_bit)
			warn = "Super", super_bit = 0;
		else if (mode_bit == alt_bit)
			warn = "Alt", alt_bit = 0;
		if (warn) {
			warned_about_overlapping_modifiers = 1;
		}
	}

	MetaMask = (meta_bit ? (1 << meta_bit) : 0);
	HyperMask = (hyper_bit ? (1 << hyper_bit) : 0);
	SuperMask = (super_bit ? (1 << super_bit) : 0);
	AltMask = (alt_bit ? (1 << alt_bit) : 0);
	ModeMask = (mode_bit ? (1 << mode_bit) : 0);	/* unused */

	if (x_modifier_keymap)
		XFreeModifiermap(x_modifier_keymap);
}


int ToolWindow::translateModifier(char *key)
{
	if (strcasecmp(key, "SHIFT") == 0 && ShiftMask != 0)
		return ShiftMask;
	else if (strcasecmp(key, "CONTROL") == 0 && ControlMask != 0)
		return ControlMask;
	else if (strcasecmp(key, "ALT") == 0 && AltMask != 0)
		return AltMask;
	else if (strcasecmp(key, "META") == 0 && MetaMask != 0)
		return MetaMask;
	else if (strcasecmp(key, "SUPER") == 0 && SuperMask != 0)
		return SuperMask;
	else if (strcasecmp(key, "HYPER") == 0 && HyperMask != 0)
		return HyperMask;
	else if (strcasecmp(key, "MOD1") == 0 && Mod1Mask != 0)
		return Mod1Mask;
	else if (strcasecmp(key, "MOD2") == 0 && Mod2Mask != 0)
		return Mod2Mask;
	else if (strcasecmp(key, "MOD3") == 0 && Mod3Mask != 0)
		return Mod3Mask;
	else if (strcasecmp(key, "MOD4") == 0 && Mod4Mask != 0)
		return Mod4Mask;
	else if (strcasecmp(key, "MOD5") == 0 && Mod5Mask != 0)
		return Mod5Mask;
	else if (strcasecmp(key, "NONE") == 0)
		return None;
	else
		return -1;
}

/* Wrapper so that we may fit our naming conventions, yet leave the 
   original XEmacs function name in place. */
void ToolWindow::InitializeModifiers(void)
{
	x_reset_modifier_mapping(getXDisplay());
}

/*-----------------------------------------------------------------------*/

/*-- end of shameless rip.  thank you for your support                ---*/

/*-----------------------------------------------------------------------*/

void ToolWindow::CheckConfig()
{
	struct stat file_status;

	if (stat(bbkeys_rcfile, &file_status) != 0) {
#ifdef DEBUG
		fprintf(stderr, "Could not open config file ->%s<-\n", bbkeys_rcfile);
#endif
	} else if (file_status.st_mtime != bbkeys_rcTime) {
		bbkeys_rcTime = file_status.st_mtime;
		loadKeygrabs();
		activateKeygrabs();
	}
}

void ToolWindow::activateKeygrabs(void)
{
	int i = 0;

	if (grabSet.instructCount > 0)
		XUngrabKey(getXDisplay(), AnyKey, AnyModifier,
							 getScreenInfo(0)->getRootWindow());
	for (i = 0; i < grabSet.instructCount; i++) {
		if (grabSet.KeyMap[i].keycode == 0)
			continue;

		if (grabSet.KeyMap[i].keycode != AnyModifier) {
			XGrabKey(getXDisplay(), grabSet.KeyMap[i].keycode,
							 grabSet.KeyMap[i].modMask | LockMask,
							 getScreenInfo(0)->getRootWindow(), True,
							 GrabModeAsync, GrabModeAsync);

			/* Also grab all modifier combinations possible that include,
			 * LockMask, ScrollLockMask and NumLockMask, so that keygrabs
			 * work even if the NumLock/ScrollLock key is on.
			 */

			wHackedGrabKey(grabSet.KeyMap[i].keycode,
										 grabSet.KeyMap[i].modMask,
										 getScreenInfo(0)->getRootWindow(), True,
										 GrabModeAsync, GrabModeAsync);
		}

		XGrabKey(getXDisplay(), grabSet.KeyMap[i].keycode,
						 grabSet.KeyMap[i].modMask,
						 getScreenInfo(0)->getRootWindow(), True, GrabModeAsync,
						 GrabModeAsync);
	}
}

/*---------------------------------------------------------------------------*/

/*--  Snarfed shamelessly from WindowMaker's code-base once again....      --*/

/*---------------------------------------------------------------------------*/
void ToolWindow::wHackedGrabKey(int keycode, unsigned int modifiers,
																Window grab_window, Bool owner_events,
																int pointer_mode, int keyboard_mode)
{
	if (modifiers == AnyModifier)
		return;
	if (keycode == AnyKey)
		return;

	/* grab all combinations of the modifier with CapsLock, NumLock and
	 * ScrollLock. How much memory/CPU does such a monstrosity consume
	 * in the server?
	 */
	if (_NumLockMask) {
		XGrabKey(getXDisplay(), keycode, modifiers | _NumLockMask,
						 grab_window, owner_events, pointer_mode, keyboard_mode);
	}
	if (_ScrollLockMask) {
		XGrabKey(getXDisplay(), keycode, modifiers | _ScrollLockMask,
						 grab_window, owner_events, pointer_mode, keyboard_mode);
	}
	if (_NumLockMask && _ScrollLockMask) {
		XGrabKey(getXDisplay(), keycode,
						 modifiers | _NumLockMask | _ScrollLockMask, grab_window,
						 owner_events, pointer_mode, keyboard_mode);
	}
	if (_NumLockMask) {
		XGrabKey(getXDisplay(), keycode,
						 modifiers | _NumLockMask | LockMask, grab_window,
						 owner_events, pointer_mode, keyboard_mode);
	}
	if (_ScrollLockMask) {
		XGrabKey(getXDisplay(), keycode,
						 modifiers | _ScrollLockMask | LockMask, grab_window,
						 owner_events, pointer_mode, keyboard_mode);
	}
	if (_NumLockMask && _ScrollLockMask) {
		XGrabKey(getXDisplay(), keycode,
						 modifiers | _NumLockMask | _ScrollLockMask | LockMask,
						 grab_window, owner_events, pointer_mode, keyboard_mode);
	}
	/* phew, I guess that's all, right? */
}

void ToolWindow::getOffendingModifiers(void)
{
	int i;
	XModifierKeymap *modmap;
	KeyCode nlock, slock;
	static int mask_table[8] = {
		ShiftMask, LockMask, ControlMask, Mod1Mask,
		Mod2Mask, Mod3Mask, Mod4Mask, Mod5Mask
	};

	nlock = XKeysymToKeycode(getXDisplay(), XK_Num_Lock);
	slock = XKeysymToKeycode(getXDisplay(), XK_Scroll_Lock);

	/*
	 * Find out the masks for the NumLock and ScrollLock modifiers,
	 * so that we can bind the grabs for when they are enabled too.
	 */
	modmap = XGetModifierMapping(getXDisplay());

	if (modmap != NULL && modmap->max_keypermod > 0) {
		for (i = 0; i < 8 * modmap->max_keypermod; i++) {
			if (modmap->modifiermap[i] == nlock && nlock != 0)
				_NumLockMask = mask_table[i / modmap->max_keypermod];
			else if (modmap->modifiermap[i] == slock && slock != 0)
				_ScrollLockMask = mask_table[i / modmap->max_keypermod];
		}
	}

	if (modmap)
		XFreeModifiermap(modmap);
}

/*-------------------------------------------------------------------------*/

/*--  End of shameful code-snarf.....................................    --*/

/*-------------------------------------------------------------------------*/

int ToolWindow::translateAction(char *action)
{
	if (!action)
		return 0;

	if (!strcasecmp(action, "Minimize"))
		return grabIconify;
	if (!strcasecmp(action, "Raise"))
		return grabRaise;
	if (!strcasecmp(action, "Lower"))
		return grabLower;
	if (!strcasecmp(action, "Close"))
		return grabClose;
	if (!strcasecmp(action, "Workspace1"))
		return grabWorkspace1;
	if (!strcasecmp(action, "Workspace2"))
		return grabWorkspace2;
	if (!strcasecmp(action, "Workspace3"))
		return grabWorkspace3;
	if (!strcasecmp(action, "Workspace4"))
		return grabWorkspace4;
	if (!strcasecmp(action, "Workspace5"))
		return grabWorkspace5;
	if (!strcasecmp(action, "Workspace6"))
		return grabWorkspace6;
	if (!strcasecmp(action, "Workspace7"))
		return grabWorkspace7;
	if (!strcasecmp(action, "Workspace8"))
		return grabWorkspace8;
	if (!strcasecmp(action, "Workspace9"))
		return grabWorkspace9;
	if (!strcasecmp(action, "Workspace10"))
		return grabWorkspace10;
	if (!strcasecmp(action, "Workspace11"))
		return grabWorkspace11;
	if (!strcasecmp(action, "Workspace12"))
		return grabWorkspace12;
	if (!strcasecmp(action, "NextWorkspace"))
		return grabNextWorkspace;
	if (!strcasecmp(action, "PrevWorkspace"))
		return grabPrevWorkspace;
	if (!strcasecmp(action, "NextWindow"))
		return grabNextWindow;
	if (!strcasecmp(action, "PrevWindow"))
		return grabPrevWindow;
	if (!strcasecmp(action, "ShadeWindow"))
		return grabShade;
	if (!strcasecmp(action, "MaximizeWindow"))
		return grabMaximize;
	if (!strcasecmp(action, "StickWindow"))
		return grabStick;
	if (!strcasecmp(action, "ExecCommand"))
		return grabExecute;
	if (!strcasecmp(action, "MaximizeVertical"))
		return grabVertMax;
	if (!strcasecmp(action, "MaximizeHorizontal"))
		return grabHorizMax;


	return 0;
}

void ToolWindow::execCommand(char *ptrCommand)
{
	int pid;
	extern char **environ;

	pid = fork();
	if (pid == -1) {
		fprintf(stderr, "bbkeys: Could not fork a process for execCommand.\n");
		return;
	}

	if (pid == 0) {
		char *argv[4];
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = ptrCommand;
		argv[3] = 0;
		execve("/bin/sh", argv, environ);
		exit(127);
	}
}

void ToolWindow::setKeygrabs(void)
{
	int pid, status;
	char command[80];
	extern char **environ;

	memset(command, 0, 80);

	if (noQt) {
		sprintf(command, "xterm -bg black -fg green -e bbkeysConfigC");
	} else {
		sprintf(command, "bbkeysconf");
	}

	pid = fork();
	if (pid == -1) {
		fprintf(stderr,
						"bbkeys: Couldn't fork a process to launch configurator.\n");
		return;
	}

	if (pid == 0) {
		char *argv[4];
		argv[0] = "sh";
		argv[1] = "-c";
		argv[2] = command;
		argv[3] = 0;
		execve("/bin/sh", argv, environ);
		exit(127);
	}

	do {
		if (waitpid(pid, &status, 0) == -1) {
			if (errno != EINTR)
				return;
		} else
			return;
	} while (1);
}

void ToolWindow::loadKeygrabs(void)
{
	// free up pointers that we get with strdup for execCommand....
	int i=0;
	for (i=0; i< grabSet.instructCount; i++) {
	   if ( grabSet.KeyMap[i].execCommand != NULL) {
	     free(grabSet.KeyMap[i].execCommand);
	  }
	}

	/* re-initialize our grabSet count... */
	grabSet.instructCount = 0;

	int count = 0;
	char line[1024];
	char Keytograb[80];
	char Modifier[80];
	char action[80];
	char execCommand[500];
	char *KeytograbBegin;
	char *ModifierBegin;
	char *actionBegin;
	char *execCmdBegin;

	FILE *rc_file = fopen(bbkeys_rcfile, "r");
	if (!rc_file) {
		/* fprintf(stderr, "Can't open bbkeys rc-file (%s)\n", bbkeys_rcfile); */
		return;
	}

	memset(line, '\0', 1024);

	while (fgets(line, 1024, rc_file) && !feof(rc_file) &&
				 count < MaxInstructions) {
		if (line[0] != '#' && strstr(line, "WithModifier")) {

			memset(Keytograb, '\0', 80);
			memset(Modifier, '\0', 80);
			memset(action, '\0', 80);
			memset(execCommand, '\0', 500);

			KeytograbBegin = strchr(line, '(');
			ModifierBegin = strchr(KeytograbBegin + 1, '(');
			actionBegin = strchr(ModifierBegin + 1, '(');

			strncat(Keytograb, KeytograbBegin + 1,
							strcspn(KeytograbBegin + 1, ")"));
			strncpy(Modifier, ModifierBegin + 1,
							strcspn(ModifierBegin + 1, ")"));
			strncpy(action, actionBegin + 1, strcspn(actionBegin + 1, ")"));

			grabSet.KeyMap[count].keycode = XKeysymToKeycode(getXDisplay(),
																											 XStringToKeysym
																											 (Keytograb));

			char *k;
			char *tmp = (char *)Modifier;
			grabSet.KeyMap[count].modMask = 0;
			while ((k = strchr(tmp, '+')) != NULL) {
				*k = 0;
				grabSet.KeyMap[count].modMask |= translateModifier(tmp);
				tmp = k + 1;
			}
			grabSet.KeyMap[count].modMask |= translateModifier(tmp);

			grabSet.KeyMap[count].action = translateAction(action);

			/* if we're supposed to having an execCommand and we do have
			 * something to put into it                                 */
			if (grabSet.KeyMap[count].action == grabExecute) {
				execCmdBegin = strchr(actionBegin + 1, '(');
				if (execCmdBegin) {
					strncpy(execCommand, execCmdBegin + 1,
									strcspn(execCmdBegin + 1, ")"));
					grabSet.KeyMap[count].execCommand = strdup(execCommand);
				}
			} else {
				grabSet.KeyMap[count].execCommand = NULL;
			}

#ifdef DEBUG
			fprintf(stdout, "Keysym ->%d<-, Modifier ->0x%lx<-, "
							"action ->%d<-, execCommand ->%s<-\n",
							grabSet.KeyMap[count].keycode,
							grabSet.KeyMap[count].modMask,
							grabSet.KeyMap[count].action,
							grabSet.KeyMap[count].execCommand);
#endif
			count++;

		}
	}
	fclose(rc_file);
	grabSet.instructCount = count;
#ifdef DEBUG
	fprintf(stderr, "loaded ->%d<- instructions.\n", count);
#endif
}

ToolWindow::ToolWindow(int argc, char **argv,
											 struct CMDOPTIONS *options):Basewindow(argc, argv,
																															options),
TimeoutHandler()
{

	timer = new BTimer(this, this);
	timer->setTimeout(100);
	timer->fireOnce(True);
	char *homedir = getenv("HOME");
	bbkeys_rcfile = new char[strlen(homedir) + 32];
	sprintf(bbkeys_rcfile, "%s/.bbkeysrc", homedir);

	XrmInitialize();

	InitializeModifiers();

	_NumLockMask = 0;
	_ScrollLockMask = 0;

	/*
	 * Ignore CapsLock in modifiers
	 */
	ValidModMask = 0xff & ~LockMask;

	getOffendingModifiers();
	/*
	 * Ignore NumLock and ScrollLock too
	 */
	ValidModMask &= ~(_NumLockMask | _ScrollLockMask);

	// initialize it so we don't run into problems later
	grabSet.instructCount = 0;
	loadKeygrabs();
	activateKeygrabs();

	iargc = argc;
	iargv = argv;

	resource = new Resource(this);
	wminterface = new WMInterface(this);
	windowList = new LinkedList < WindowList >;
	desktopList = new LinkedList < DesktopList >;

	desktop_nr = 0;
	current_desktop_nr = -1;
	focus_window = None;
	wm_init = False;
	number_of_desktops = 0;
	row_last = column_last = 0;
	MakeWindow(False);
	Redraw();
	wminterface->moduleInit();

	eventLoop();

}

ToolWindow::~ToolWindow()
{
	XUnmapWindow(getXDisplay(), framewin);

	/* destroy pixmaps */
	if (pixmap.frame)
		getImageControl()->removeImage(pixmap.frame);
	if (pixmap.pix_title)
		getImageControl()->removeImage(pixmap.pix_title);
	if (pixmap.pix_back)
		getImageControl()->removeImage(pixmap.pix_back);
	if (pixmap.pix_configBtn)
		getImageControl()->removeImage(pixmap.pix_configBtn);
	if (pixmap.pix_closeBtn)
		getImageControl()->removeImage(pixmap.pix_closeBtn);
	if (pixmap.pix_pressedBtn)
		getImageControl()->removeImage(pixmap.pix_pressedBtn);

	/* destroy windows */
	XDestroyWindow(getXDisplay(), framewin);
	XDestroyWindow(getXDisplay(), win_back);
	XDestroyWindow(getXDisplay(), win_title);
	XDestroyWindow(getXDisplay(), win_configBtn);
	XDestroyWindow(getXDisplay(), win_closeBtn);

	/* destroy lists */
	delete windowList;
	delete desktopList;
}

void ToolWindow::CycleWindowFocus(bool focus_next_window)
{
	unsigned long int data = 0;
	if (focus_next_window) {
		data = 1;
	} else {
		data = 0;
	}

	wminterface->sendClientMessage(getBlackboxCycleWindowFocusAtom(), data);

}

void ToolWindow::reconfigure(void)
{

	/* destroy pixmaps */
	if (pixmap.frame)
		getImageControl()->removeImage(pixmap.frame);
	if (pixmap.pix_title)
		getImageControl()->removeImage(pixmap.pix_title);
	if (pixmap.pix_back)
		getImageControl()->removeImage(pixmap.pix_back);
	if (pixmap.pix_configBtn)
		getImageControl()->removeImage(pixmap.pix_configBtn);
	if (pixmap.pix_closeBtn)
		getImageControl()->removeImage(pixmap.pix_closeBtn);
	if (pixmap.pix_pressedBtn)
		getImageControl()->removeImage(pixmap.pix_pressedBtn);

	resource->Reload();

	MakeWindow(True);

	LinkedListIterator < DesktopList > desktop_it(desktopList);
	LinkedListIterator < WindowList > win_it(windowList);

	XClearWindow(getXDisplay(), framewin);
	XClearWindow(getXDisplay(), win_back);
	Redraw();

}

void ToolWindow::MakeWindow(bool reconfigure)
{
	XSetWindowAttributes attrib;
	XWMHints wmhints;
	XClassHint classhints;
	XTextProperty windowname;

	unsigned long create_mask = CWBackPixmap | CWBorderPixel |
		 CWOverrideRedirect | CWCursor | CWEventMask;

	// geometry for our windows...................................

	geom_title.height = resource->label.font->ascent +
		 resource->label.font->descent + 2;
	if (geom_title.height < 20 && (miniMe || tinyMe))
		geom_title.height = 20;
	geom_title.width = XTextWidth(resource->label.font, "bbkeys",
																strlen("bbkeys")) + (geom_title.height +
																										 5);
	geom_title.x = 2;
	geom_title.y = 2;
	geom_back.width = geom_title.width + 4;
	geom_back.x = 0;
	geom_back.y = 0;
	geom_closeBtn.height = geom_title.height * 7 / 10;
	geom_closeBtn.width = geom_closeBtn.height;

	if (miniMe) {

		geom_back.height = geom_title.height + 2;
		geom_closeBtn.y = 80;				/* put it out of sight. */
		geom_closeBtn.x = 80;
		geom_configBtn.height = geom_title.height - 1;
		geom_configBtn.width = geom_configBtn.height;
		geom_configBtn.x = geom_title.width - geom_configBtn.width;
		geom_configBtn.y = geom_title.y;

	} else if (tinyMe) {

		geom_title.height = 22;
		geom_title.width = 22;
		geom_title.x = 80;
		geom_title.y = 80;
		geom_back.width = 22;
		geom_back.height = 22;
		geom_closeBtn.y = 80;				/* put it out of sight. */
		geom_closeBtn.x = 80;
		geom_configBtn.height = 20;
		geom_configBtn.width = geom_configBtn.height;
		geom_configBtn.x = 1;
		geom_configBtn.y = 1;

	} else {

		geom_back.height = 3 * (geom_title.height) + 6;
		geom_closeBtn.x = geom_title.width - geom_closeBtn.width - 1;
		geom_closeBtn.y = geom_title.y +
			 ((geom_title.height - geom_closeBtn.height) / 2);
		geom_configBtn.height = (geom_title.height * 2);
		geom_configBtn.width = geom_title.width - 4;
		geom_configBtn.x = geom_title.x + 2;
		geom_configBtn.y = geom_title.height + geom_title.y + 2;

	}

	frame.height = geom_back.height;
	frame.width = geom_back.width;
	frame.x = resource->position.x;
	frame.y = resource->position.y;

	// end geometry for our windows...................................

	if (resource->position.mask & XNegative) {
		frame.x = getCurrentScreenInfo()->getWidth() +
			 resource->position.x - frame.width;
	}

	if (resource->position.mask & YNegative) {
		frame.y = getCurrentScreenInfo()->getHeight() +
			 resource->position.y - frame.height;

	}

	if (withdrawn) {
		attrib.override_redirect = False;
		wmhints.initial_state = WithdrawnState;
	} else if (iconic) {
		attrib.override_redirect = False;
		wmhints.initial_state = IconicState;
	} else {
		attrib.override_redirect = False;
		wmhints.initial_state = NormalState;
	}

	attrib.background_pixmap = ParentRelative;

	pixmap.frame =
		 getImageControl()->renderImage(frame.width, frame.height,
																		&resource->frame.texture);

	pixmap.pix_back =
		 getImageControl()->renderImage(geom_back.width,
																		geom_back.height,
																		&resource->frame.texture);
	pixmap.pix_title =
		 getImageControl()->renderImage(geom_title.width, geom_title.height,
																		&resource->label.texture);

	pixmap.pix_configBtn =
		 getImageControl()->renderImage(geom_configBtn.width,
																		geom_configBtn.height,
																		&resource->button.texture);

	pixmap.pix_closeBtn =
		 getImageControl()->renderImage(geom_closeBtn.width,
																		geom_closeBtn.height,
																		&resource->button.texture);
	pixmap.pix_pressedBtn =
		 getImageControl()->renderImage(geom_configBtn.width,
																		geom_configBtn.height,
																		&resource->button.texture_pressed);

	attrib.cursor = getSessionCursor();
	attrib.event_mask =
		 ButtonPressMask | ButtonReleaseMask | ExposureMask |
		 FocusChangeMask | KeyPressMask | StructureNotifyMask |
		 SubstructureRedirectMask;

	if (!reconfigure) {
		framewin = XCreateWindow(getXDisplay(),
														 getCurrentScreenInfo()->getRootWindow(),
														 frame.x, frame.y, frame.width,
														 frame.height, 0,
														 getCurrentScreenInfo()->getDepth(),
														 InputOutput,
														 getCurrentScreenInfo()->getVisual(),
														 create_mask, &attrib);
		win_back = XCreateWindow(getXDisplay(),
														 framewin,
														 geom_back.x,
														 geom_back.y,
														 geom_back.width,
														 geom_back.height, 0,
														 getCurrentScreenInfo
														 ()->getDepth(), InputOutput,
														 getCurrentScreenInfo
														 ()->getVisual(), create_mask, &attrib);
		win_title =
			 XCreateWindow(getXDisplay(), framewin, geom_title.x,
										 geom_title.y, geom_title.width,
										 geom_title.height, 0,
										 getCurrentScreenInfo()->getDepth(), InputOutput,
										 getCurrentScreenInfo()->getVisual(), create_mask,
										 &attrib);
		win_configBtn =
			 XCreateWindow(getXDisplay(), framewin, geom_configBtn.x,
										 geom_configBtn.y, geom_configBtn.width,
										 geom_configBtn.height, 0,
										 getCurrentScreenInfo()->getDepth(), InputOutput,
										 getCurrentScreenInfo()->getVisual(), create_mask,
										 &attrib);
		win_closeBtn =
			 XCreateWindow(getXDisplay(), framewin, geom_closeBtn.x,
										 geom_closeBtn.y, geom_closeBtn.width,
										 geom_closeBtn.height, 0,
										 getCurrentScreenInfo()->getDepth(), InputOutput,
										 getCurrentScreenInfo()->getVisual(), create_mask,
										 &attrib);
	} else if (!withdrawn) {
		XMoveResizeWindow(getXDisplay(), framewin, frame.x, frame.y,
											frame.width, frame.height);
		XMoveResizeWindow(getXDisplay(), win_back,
											geom_back.x, geom_back.y,
											geom_back.width, geom_back.height);
		XMoveResizeWindow(getXDisplay(), win_title, geom_title.x,
											geom_title.y, geom_title.width, geom_title.height);
		XMoveResizeWindow(getXDisplay(), win_configBtn, geom_configBtn.x,
											geom_configBtn.y, geom_configBtn.width,
											geom_configBtn.height);
		XMoveResizeWindow(getXDisplay(), win_closeBtn, geom_closeBtn.x,
											geom_closeBtn.y, geom_closeBtn.width,
											geom_closeBtn.height);

	} else {
		XResizeWindow(getXDisplay(), framewin, frame.width, frame.height);
		XResizeWindow(getXDisplay(), win_back,
									geom_back.width, geom_back.height);
		XResizeWindow(getXDisplay(), win_title, geom_title.width,
									geom_title.height);
		XMoveResizeWindow(getXDisplay(), win_configBtn, geom_configBtn.x,
											geom_configBtn.y, geom_configBtn.width,
											geom_configBtn.height);
		XMoveResizeWindow(getXDisplay(), win_closeBtn, geom_closeBtn.x,
											geom_closeBtn.y, geom_closeBtn.width,
											geom_closeBtn.height);
	}

	char *name = BBTOOL;
	XSizeHints sizehints;

	wmhints.flags = StateHint;

	classhints.res_name = BBTOOL;
	classhints.res_class = "bbtools";

	sizehints.x = frame.x;				//getResource()->position.x;
	sizehints.y = frame.y;				//getResource()->position.y;

	sizehints.max_width = sizehints.min_width = frame.width;
	sizehints.max_height = sizehints.min_height = frame.height;
	sizehints.flags = USPosition | PMinSize | PMaxSize;

	XStringListToTextProperty(&name, 1, &windowname);
	XSetWMProperties(getXDisplay(), framewin, &windowname, NULL, getArgv(),
									 getArgc(), &sizehints, &wmhints, &classhints);

	// free up allocated memory in XStringListToTextProperty
	XFree(windowname.value);

	if (!reconfigure) {
	   Atom wmproto[2];
	   wmproto[0] = wm_delete_window;
	   wmproto[1] = getBlackboxStructureMessagesAtom();
	   XSetWMProtocols(getXDisplay(), framewin, wmproto, 2);
  }

	if (!decorated && !withdrawn) {
		BlackboxHints net_hints;
		net_hints.decoration = DecorNone;
		net_hints.attrib = AttribOmnipresent;
		net_hints.flags = AttribDecoration | AttribOmnipresent;
		XChangeProperty(getXDisplay(), framewin, getBlackboxHintsAtom(),
										getBlackboxHintsAtom(), 32, PropModeReplace,
										(unsigned char *) &net_hints,
										PropBlackboxHintsElements);
	}

	if (!shape) {
		XSetWindowBackgroundPixmap(getXDisplay(), framewin, pixmap.frame);
	}

	XSetWindowBackgroundPixmap(getXDisplay(), win_title, pixmap.pix_title);
	XSetWindowBackgroundPixmap(getXDisplay(), win_back, pixmap.pix_back);
	XSetWindowBackgroundPixmap(getXDisplay(), win_configBtn,
														 pixmap.pix_configBtn);
	XSetWindowBackgroundPixmap(getXDisplay(), win_closeBtn,
														 pixmap.pix_closeBtn);


	if (!withdrawn && resource->report.auto_raise) {
		XRaiseWindow(getXDisplay(), framewin);
		lower = False;
	} else
		lower = True;

	if (!reconfigure) {
		gcv.font = resource->label.font->fid;
		gcv.foreground = resource->label.textColor.getPixel();
		frameGC =
			 XCreateGC(getXDisplay(), framewin, GCFont | GCForeground, &gcv);
	} else {
		gcv.font = resource->label.font->fid;
		gcv.foreground = resource->label.textColor.getPixel();
		XChangeGC(getXDisplay(), frameGC, GCFont | GCForeground, &gcv);
	}


	if (!reconfigure) {
		XClearWindow(getXDisplay(), framewin);
		XMapWindow(getXDisplay(), framewin);
		XMapSubwindows(getXDisplay(), framewin);
		XSetIconName(getXDisplay(), framewin, "bbkeys");
	}

}

void ToolWindow::Redraw()
{
	XClearWindow(getXDisplay(), win_title);
	XClearWindow(getXDisplay(), win_configBtn);
	XClearWindow(getXDisplay(), win_closeBtn);

	char *title = "bbkeys";
	XDrawString(getXDisplay(), win_title, frameGC, geom_title.x - 1,
							(geom_title.height + resource->label.font->ascent -
							 resource->label.font->descent) / 2, title, strlen(title));

	XPoint pts[3];

	if (miniMe || tinyMe) {
		pts[0].x = (geom_configBtn.width / 2);
		pts[0].y = (geom_configBtn.height / 2) - 2;
		pts[1].x = -(geom_configBtn.width / 4);
		pts[1].y = (geom_configBtn.height - pts[0].y) - 1;
		pts[2].x = (geom_configBtn.width / 2);
		pts[2].y = 0;

		XFillPolygon(getXDisplay(), win_configBtn, frameGC, pts, 3,
								 Convex, CoordModePrevious);
		XFillArc(getXDisplay(), win_configBtn, frameGC,
						 geom_configBtn.width / 2 - geom_configBtn.width / 4, 1,	// x, y
						 geom_configBtn.width / 2, geom_configBtn.height / 2,	//width,height
						 0, 360 * 64);
	} else {
		pts[0].x = (geom_configBtn.width / 2);
		pts[0].y = (geom_configBtn.height / 2) - 2;
		pts[1].x = -(geom_configBtn.width / 4);
		pts[1].y = (geom_configBtn.height - pts[0].y) - 2;
		pts[2].x = (geom_configBtn.width / 2);
		pts[2].y = 0;

		XFillPolygon(getXDisplay(), win_configBtn, frameGC, pts, 3,
								 Convex, CoordModePrevious);
		XFillArc(getXDisplay(), win_configBtn, frameGC,
						 geom_configBtn.width / 2 - geom_configBtn.width / 4, 1,	// x, y
						 geom_configBtn.width / 2, geom_configBtn.height * 5 / 8,	//width,height
						 0, 360 * 64);
	}

	/* Our little close button (drawn twice for a nice thickness) */
	XDrawLine(getXDisplay(), win_closeBtn, frameGC, 1, 1,
						geom_closeBtn.width - 2, geom_closeBtn.height - 1);
	XDrawLine(getXDisplay(), win_closeBtn, frameGC, 1,
						geom_closeBtn.height - 1, geom_closeBtn.width - 2, 1);

	XDrawLine(getXDisplay(), win_closeBtn, frameGC, 2, 1,
						geom_closeBtn.width - 1, geom_closeBtn.height - 1);
	XDrawLine(getXDisplay(), win_closeBtn, frameGC, 2,
						geom_closeBtn.height - 1, geom_closeBtn.width - 1, 1);

}

void ToolWindow::process_event(XEvent * e)
{
	switch (e->type) {
	case KeyPress:
		{
			int i = 0;
			int grabInt = -1;

			/* Need to take out this next bit when nyz has 0.60.0 ready, as
			   we'll then get our focus_window in a ClientMessage */
			int revert_to = 0;
			if (!focus_window)
				XGetInputFocus(getXDisplay(), &focus_window, &revert_to);

			/* end temporary focus_window fix */

			// if our user wants to grab his keystrokes, even though one
			// or more of the Lock-Modifiers are pressed, alter the mask
			// to make this possible....

			if (!honor_modifiers) {
				e->xkey.state = e->xkey.state & ~_NumLockMask & ~_ScrollLockMask &
					 ~LockMask;
			}

			for (i = 0; i < grabSet.instructCount; i++) {
				if ((e->xkey.keycode == grabSet.KeyMap[i].keycode) &&
						(e->xkey.state == grabSet.KeyMap[i].modMask)) {
					grabInt = i;
					break;
				}
			}

			if (grabInt > -1) {
				/* play with colors for nyz =:) */
				XSetWindowBackgroundPixmap(getXDisplay(), win_configBtn,
																	 pixmap.pix_pressedBtn);
				Redraw();

				switch (grabSet.KeyMap[grabInt].action) {
				case grabIconify:
					if (focus_window)
						XIconifyWindow(getXDisplay(), focus_window, 0);
					break;

				case grabRaise:
					if (focus_window)
						XRaiseWindow(getXDisplay(), focus_window);
					break;

				case grabLower:
					if (focus_window)
						XLowerWindow(getXDisplay(), focus_window);
					break;

				case grabClose:
					if (!focus_window)
						break;

					XEvent ce;

					ce.xclient.type = ClientMessage;
					ce.xclient.message_type = getWMProtocolsAtom();
					ce.xclient.display = getXDisplay();
					ce.xclient.window = focus_window;
					ce.xclient.format = 32;
					ce.xclient.data.l[0] = getWMDeleteAtom();
					ce.xclient.data.l[1] = CurrentTime;
					ce.xclient.data.l[2] = ce.xclient.data.l[3] =
						 ce.xclient.data.l[4] = 0l;

					XSendEvent(getXDisplay(), focus_window, False, NoEventMask, &ce);

					break;

				case grabWorkspace1:
					wminterface->changeDesktop(0);
					break;

				case grabWorkspace2:
					wminterface->changeDesktop(1);
					break;

				case grabWorkspace3:
					wminterface->changeDesktop(2);
					break;

				case grabWorkspace4:
					wminterface->changeDesktop(3);
					break;

				case grabWorkspace5:
					wminterface->changeDesktop(4);
					break;

				case grabWorkspace6:
					wminterface->changeDesktop(5);
					break;

				case grabWorkspace7:
					wminterface->changeDesktop(6);
					break;

				case grabWorkspace8:
					wminterface->changeDesktop(7);
					break;

				case grabWorkspace9:
					wminterface->changeDesktop(8);
					break;

				case grabWorkspace10:
					wminterface->changeDesktop(9);
					break;

				case grabWorkspace11:
					wminterface->changeDesktop(10);
					break;

				case grabWorkspace12:
					wminterface->changeDesktop(11);
					break;

				case grabNextWorkspace:
					if ((current_desktop_nr < number_of_desktops - 1) &&
							(current_desktop_nr >= 0))
						wminterface->changeDesktop(current_desktop_nr + 1);
					else
						wminterface->changeDesktop(0);

					break;

				case grabPrevWorkspace:
					if (current_desktop_nr > 0)
						wminterface->changeDesktop(current_desktop_nr - 1);
					else
						wminterface->changeDesktop(number_of_desktops - 1);

					break;

				case grabNextWindow:
					CycleWindowFocus(True);
					break;

				case grabPrevWindow:
					CycleWindowFocus(False);
					break;

				case grabShade:{
						if (!focus_window)
							break;

						XEvent e;
						e.xclient.type = ClientMessage;
						e.xclient.message_type = getBlackboxChangeAttributesAtom();
						e.xclient.window = focus_window;
						e.xclient.format = 32;
						e.xclient.data.l[0] = AttribShaded;
						e.xclient.data.l[2] = e.xclient.data.l[3] =
							 e.xclient.data.l[4] = 0l;

						Atom atom_return;
						int foo;
						unsigned long ulfoo, nitems;
						BlackboxAttributes *net;
						if (XGetWindowProperty(getXDisplay(), focus_window,
																	 getBlackboxAttributesAtom(), 0l,
																	 PropBlackboxAttributesElements,
																	 False,
																	 getBlackboxAttributesAtom(),
																	 &atom_return, &foo, &nitems,
																	 &ulfoo,
																	 (unsigned char **) &net) ==
								Success && net && nitems == PropBlackboxAttributesElements) {
							e.xclient.data.l[1] = net->attrib ^ AttribShaded;

							XFree((void *) net);
						} else
							e.xclient.data.l[1] = AttribShaded;

						XSendEvent(getXDisplay(),
											 getScreenInfo(0)->getRootWindow(),
											 False, SubstructureRedirectMask, &e);

						break;
					}

				case grabMaximize:{
						if (!focus_window)
							break;

						XEvent e;
						e.xclient.type = ClientMessage;
						e.xclient.message_type = getBlackboxChangeAttributesAtom();
						e.xclient.window = focus_window;
						e.xclient.format = 32;
						e.xclient.data.l[0] = AttribMaxHoriz | AttribMaxVert;
						e.xclient.data.l[2] = e.xclient.data.l[3] =
							 e.xclient.data.l[4] = 0l;

						Atom atom_return;
						int foo;
						unsigned long ulfoo, nitems;
						BlackboxAttributes *net;
						if (XGetWindowProperty(getXDisplay(), focus_window,
																	 getBlackboxAttributesAtom(), 0l,
																	 PropBlackboxAttributesElements,
																	 False,
																	 getBlackboxAttributesAtom(),
																	 &atom_return, &foo, &nitems,
																	 &ulfoo,
																	 (unsigned char **) &net) ==
								Success && net && nitems == PropBlackboxAttributesElements) {
							e.xclient.data.l[1] =
								 net->attrib ^ (AttribMaxHoriz | AttribMaxVert);

							XFree((void *) net);
						} else
							e.xclient.data.l[1] = (AttribMaxHoriz | AttribMaxVert);

						XSendEvent(getXDisplay(),
											 getScreenInfo(0)->getRootWindow(),
											 False, SubstructureRedirectMask, &e);

						break;
					}

				case grabStick:{
						if (!focus_window)
							break;

						XEvent e;
						e.xclient.type = ClientMessage;
						e.xclient.message_type = getBlackboxChangeAttributesAtom();
						e.xclient.window = focus_window;
						e.xclient.format = 32;
						e.xclient.data.l[0] = AttribOmnipresent;
						e.xclient.data.l[2] = e.xclient.data.l[3] =
							 e.xclient.data.l[4] = 0l;

						Atom atom_return;
						int foo;
						unsigned long ulfoo, nitems;
						BlackboxAttributes *net;
						if (XGetWindowProperty(getXDisplay(), focus_window,
																	 getBlackboxAttributesAtom(), 0l,
																	 PropBlackboxAttributesElements,
																	 False,
																	 getBlackboxAttributesAtom(),
																	 &atom_return, &foo, &nitems,
																	 &ulfoo,
																	 (unsigned char **) &net) ==
								Success && net && nitems == PropBlackboxAttributesElements) {
							e.xclient.data.l[1] = net->attrib ^ AttribOmnipresent;

							XFree((void *) net);
						} else
							e.xclient.data.l[1] = AttribOmnipresent;

						XSendEvent(getXDisplay(),
											 getScreenInfo(0)->getRootWindow(),
											 False, SubstructureRedirectMask, &e);

						break;
					}
				case grabExecute:{
						execCommand(grabSet.KeyMap[grabInt].execCommand);
						break;
					}

				case grabVertMax:{
						if (!focus_window)
							break;

						XEvent e;
						e.xclient.type = ClientMessage;
						e.xclient.message_type = getBlackboxChangeAttributesAtom();
						e.xclient.window = focus_window;
						e.xclient.format = 32;
						e.xclient.data.l[0] = AttribMaxVert;
						e.xclient.data.l[2] = e.xclient.data.l[3] =
							 e.xclient.data.l[4] = 0l;

						Atom atom_return;
						int foo;
						unsigned long ulfoo, nitems;
						BlackboxAttributes *net;
						if (XGetWindowProperty(getXDisplay(), focus_window,
																	 getBlackboxAttributesAtom(), 0l,
																	 PropBlackboxAttributesElements,
																	 False,
																	 getBlackboxAttributesAtom(),
																	 &atom_return, &foo, &nitems,
																	 &ulfoo,
																	 (unsigned char **) &net) ==
								Success && net && nitems == PropBlackboxAttributesElements) {
							e.xclient.data.l[1] = net->attrib ^ (AttribMaxVert);

							XFree((void *) net);
						} else
							e.xclient.data.l[1] = (AttribMaxVert);

						XSendEvent(getXDisplay(),
											 getScreenInfo(0)->getRootWindow(),
											 False, SubstructureRedirectMask, &e);

						break;
					}

				case grabHorizMax:{
						if (!focus_window)
							break;

						XEvent e;
						e.xclient.type = ClientMessage;
						e.xclient.message_type = getBlackboxChangeAttributesAtom();
						e.xclient.window = focus_window;
						e.xclient.format = 32;
						e.xclient.data.l[0] = AttribMaxHoriz;
						e.xclient.data.l[2] = e.xclient.data.l[3] =
							 e.xclient.data.l[4] = 0l;

						Atom atom_return;
						int foo;
						unsigned long ulfoo, nitems;
						BlackboxAttributes *net;
						if (XGetWindowProperty(getXDisplay(), focus_window,
																	 getBlackboxAttributesAtom(), 0l,
																	 PropBlackboxAttributesElements,
																	 False,
																	 getBlackboxAttributesAtom(),
																	 &atom_return, &foo, &nitems,
																	 &ulfoo,
																	 (unsigned char **) &net) ==
								Success && net && nitems == PropBlackboxAttributesElements) {
							e.xclient.data.l[1] = net->attrib ^ (AttribMaxHoriz);

							XFree((void *) net);
						} else
							e.xclient.data.l[1] = (AttribMaxHoriz);

						XSendEvent(getXDisplay(),
											 getScreenInfo(0)->getRootWindow(),
											 False, SubstructureRedirectMask, &e);

						break;
					}
				}

				timer->start();
			}

			break;
		}

	case ClientMessage:
		{
			if ((unsigned) e->xclient.data.l[0] == wm_delete_window)
				shutdown();
			wminterface->handleNETEvents(*e);
		}

		break;

	case Expose:
		if (!e->xexpose.count) {
			reconfigure();
		}
		break;

	case ButtonPress:
		if (e->xbutton.button == LEFT_BUTTON
				|| e->xbutton.button == RIGHT_BUTTON) {
			if (e->xbutton.window == win_configBtn) {
				XSetWindowBackgroundPixmap(getXDisplay(), win_configBtn,
																	 pixmap.pix_pressedBtn);
				Redraw();
				/* Redraw(); */
			} else if (e->xbutton.window == win_closeBtn) {
				XSetWindowBackgroundPixmap(getXDisplay(), win_closeBtn,
																	 pixmap.pix_pressedBtn);
				Redraw();
			} else if (e->xbutton.window == win_title) {
				if (!raised) {
					XRaiseWindow(getXDisplay(), framewin);
					raised = True;
				}
			}
		} else if (e->xbutton.button == MIDDLE_BUTTON) {
			if (raised) {
				XLowerWindow(getXDisplay(), framewin);
				raised = False;
			}
		}

		break;

	case ButtonRelease:
		if (e->xbutton.button == LEFT_BUTTON
				|| e->xbutton.button == RIGHT_BUTTON) {
			if (e->xbutton.window == win_configBtn) {
				if ((e->xbutton.x > 0)
						&& (e->xbutton.x < geom_configBtn.width)
						&& (e->xbutton.y > 0)
						&& (e->xbutton.y < geom_configBtn.height)) {

					if (grabSet.instructCount > 0) {
						XUngrabKey(getXDisplay(), AnyKey, AnyModifier,
											 getScreenInfo(0)->getRootWindow());
						XSync(getXDisplay(), False);
					}

					setKeygrabs();

					/*   discard all XEvents (was causing us to go bananas on the
					 *   user before, since we were executing all keygrabs at once
					 *   that the user did when the configuration app was running)
					 *   *cough* undocumented feature *cough*
					 */

					XEvent event;
					while (XPending(getXDisplay())) {
						XNextEvent(getXDisplay(), &event);
					}

					/* carry on then, little soldier */

					loadKeygrabs();
					activateKeygrabs();
					timer->start();

				}

				XSetWindowBackgroundPixmap(getXDisplay(), win_configBtn,
																	 pixmap.pix_configBtn);
				Redraw();
			} else if (e->xbutton.window == win_closeBtn) {
				if ((e->xbutton.x > 0)
						&& (e->xbutton.x < geom_closeBtn.width)
						&& (e->xbutton.y > 0)
						&& (e->xbutton.y < geom_closeBtn.height)) {

					if (e->xbutton.button == LEFT_BUTTON) {
						shutdown();
					} else if (e->xbutton.button == RIGHT_BUTTON) {
						if (withdrawn) {
							XUnmapWindow(getXDisplay(), framewin);
							withdrawn = False;
							iconic = True;

							XWMHints wmhints;
							wmhints.initial_state = IconicState;
							wmhints.flags = StateHint;
							XSetWMHints(getXDisplay(), framewin, &wmhints);
							XMapWindow(getXDisplay(), framewin);

							/* this should work but doesn't seem to.  *shrug* */
							/* reconfigure(); */
						} else {
							XIconifyWindow(getXDisplay(), framewin, 0);
						}
					}
				}

				XSetWindowBackgroundPixmap(getXDisplay(), win_closeBtn,
																	 pixmap.pix_configBtn);
				Redraw();
			}
		}

		break;

	case ConfigureNotify:
		if ((e->xconfigure.window == framewin) && e->xconfigure.send_event) {
			if (withdrawn) {
				reconfigure();
			}

			int parent_x, parent_y;
			Window parent_root;
			unsigned int parent_width, parent_height, parent_border_width,
				 parent_depth;

			if (withdrawn) {
				XGetGeometry(getXDisplay(), e->xconfigure.above,
										 &parent_root, &parent_x, &parent_y,
										 &parent_width, &parent_height,
										 &parent_border_width, &parent_depth);

				frame.x = e->xconfigure.x + parent_x;
				frame.y = e->xconfigure.y + parent_y;
			}
		}

		break;
	}
}


void ToolWindow::timeout(void)
{
	XSetWindowBackgroundPixmap(getXDisplay(), win_configBtn,
														 pixmap.pix_configBtn);
	Redraw();
}

void ToolWindow::moveWinToDesktop(Window win, DesktopList * desktop)
{

	LinkedListIterator < WindowList > win_it(windowList);
	for (; win_it.current(); win_it++)
		if ((win_it.current()->win) == win)
			break;
	if (win_it.current()) {
		if ((desktop->desktop_nr != win_it.current()->desktop_nr)
				& (!win_it.current()->sticky)) {

			XUnmapWindow(getXDisplay(), win_it.current()->pager_win);
			XReparentWindow(getXDisplay(), win_it.current()->pager_win,
											desktop->win, win_it.current()->pager_x,
											win_it.current()->pager_y);
			XMapWindow(getXDisplay(), win_it.current()->pager_win);
			win_it.current()->desktop_nr = desktop_nr;
		}
	}
}

int ToolWindow::getWindowGeometry(struct WindowList *winlist)
{
	unsigned int border_width, depth;
	Window root_return, child_return;
	int x_return, y_return;
	int status;

	status =
		 XGetGeometry(getXDisplay(), winlist->win, &root_return, &x_return,
									&y_return, &winlist->width, &winlist->height,
									&border_width, &depth);
	if (status) {
		XTranslateCoordinates(getXDisplay(), winlist->win, root_return,
													x_return, y_return, &winlist->x_position,
													&winlist->y_position, &child_return);
		return 1;
	}
	return 0;
}

/*-------------------------------------*/

/*-  old bbpager code....            --*/

/*-------------------------------------*/

void ToolWindow::raiseWindow(Window win)
{
	LinkedListIterator < WindowList > win_it(windowList);
	for (; win_it.current(); win_it++)
		if ((win_it.current()->win) == win)
			break;

	if (win_it.current()) {
		WindowList *tmp = win_it.current();
		XRaiseWindow(getXDisplay(), tmp->pager_win);
	}
}

void ToolWindow::lowerWindow(Window win)
{
	LinkedListIterator < WindowList > win_it(windowList);
	for (; win_it.current(); win_it++)
		if ((win_it.current()->win) == win)
			break;

	if (win_it.current()) {
		WindowList *tmp = win_it.current();
		XLowerWindow(getXDisplay(), tmp->pager_win);
	}
}

void ToolWindow::focusWindow(Window win)
{
	focus_window = win;
}

void ToolWindow::desktopChange(int desktop_nr)
{
	current_desktop_nr = desktop_nr;
}

int ToolWindow::winOnDesktop(Window win)
{
	return (0);
}

void ToolWindow::removeWindow(Window win)
{
}

void ToolWindow::removeDesktopWindow()
{
}

void ToolWindow::addDesktopWindow(struct DesktopList *tmp,
																	bool reconfigure)
{
}

void ToolWindow::addFrameWindow(struct WindowList *window,
																Window desktopWin, bool reconfigure)
{
}
